#include "catch2/catch.hpp"

#include <csg.hpp>
#include <csg_types.hpp>
#include <csg_matrix_functions.hpp>

namespace {

TEST_CASE("linear", "[Levelset Extrude]") {
  double height = 100, radius = 10;
  auto my_lin_ext = csg::LinearExtrude{
      .height = 100, .center = false, .scale = {1, 1}, .group = csg::Union2D()};
  csg::Circle my_cir{.name = std::nullopt, .radius = 10};

  SECTION("Not centered") {
    my_lin_ext.center = false;
    my_lin_ext.group.objs.push_back(my_cir);
    auto my_tree = std::make_shared<csg::Tree>();
    my_tree->top.objs.push_back(my_lin_ext);
    csg::CsgIF my_levelset(my_tree);

    SECTION("Outside") {
      CHECK_FALSE(0 < my_levelset(0, radius * 0.99, -0.01 * height));
      CHECK_FALSE(0 < my_levelset(0, radius * 1.01, 0.01 * height));
      CHECK_FALSE(0 < my_levelset(0, radius * 0.99, 1.01 * height));
      CHECK_FALSE(0 < my_levelset(radius * 0.99, 0, 1.01 * height));
    }
    SECTION("Inside") {
      CHECK(0 < my_levelset(0, radius * 0.99, 0.01 * height));
      CHECK(0 < my_levelset(radius * 0.99, 0, 0.01 * height));
      CHECK(0 < my_levelset(0, radius * 0.99, 0.99 * height));
      CHECK(0 < my_levelset(radius * 0.99, 0, 0.99 * height));
    }
  }

  SECTION("Centered") {
    my_lin_ext.center = true;
    my_lin_ext.group.objs.push_back(my_cir);
    auto my_tree = std::make_shared<csg::Tree>();
    my_tree->top.objs.push_back(my_lin_ext);
    csg::CsgIF my_levelset(my_tree);

    SECTION("Outside") {
      CHECK_FALSE(0 < my_levelset(0, radius * 0.99, -1.01 * height / 2));
      CHECK_FALSE(0 < my_levelset(0, radius * 1.01, 0.01 * height));
      CHECK_FALSE(0 < my_levelset(0, radius * 0.99, 1.01 * height / 2));
      CHECK_FALSE(0 < my_levelset(radius * 0.99, 0, 1.01 * height / 2));
    }
    SECTION("Inside") {
      CHECK(0 < my_levelset(0, radius * 0.99, -0.99 * height / 2));
      CHECK(0 < my_levelset(radius * 0.99, 0, 0.01 * height / 2));
      CHECK(0 < my_levelset(0, radius * 0.99, 0.99 * height / 2));
      CHECK(0 < my_levelset(radius * 0.99, 0, 0.99 * height / 2));
    }
  }
}

TEST_CASE("two shape linear extrude", "[Levelset Extrude]") {
  auto my_lin_ext = csg::LinearExtrude{
      .height = 100, .center = true, .scale = {1, 1}, .group = csg::Union2D()};

  csg::Circle my_cir{.name = std::nullopt, .radius = 1};
  my_lin_ext.group.objs.push_back(my_cir);

  auto my_mat = csg::Mulmatrix2D({{1, 0}}, {{0, 1}});
  my_mat.translation = {4, 0};
  my_mat.group.objs.push_back(my_cir);
  my_lin_ext.group.objs.push_back(my_mat);

  auto my_tree = std::make_shared<csg::Tree>();
  my_tree->top.objs.push_back(my_lin_ext);
  csg::CsgIF my_levelset(my_tree);

  SECTION("Outside") {
    CHECK_FALSE(0 < my_levelset(0, 2, 0));
    CHECK_FALSE(0 < my_levelset(-4, 0, 0));
    CHECK_FALSE(0 < my_levelset(3.5, 0, 51));
    CHECK_FALSE(0 < my_levelset(3.5, 0, -51));
  }
  SECTION("Inside") {
    CHECK(0 < my_levelset(0, 0, 0));
    CHECK(0 < my_levelset(0.5, 0, 0));
    CHECK(0 < my_levelset(-0.5, 0, 0));
    CHECK(0 < my_levelset(4.5, 0, 0));
    CHECK(0 < my_levelset(3.5, 0, 0));
    CHECK(0 < my_levelset(3.5, 0, 49));
    CHECK(0 < my_levelset(3.5, 0, -49));
  }
}

TEST_CASE("simple torus", "[Levelset Extrude]") {
  auto my_rot_ext = csg::RotateExtrude();

  csg::Circle my_cir{.name = std::nullopt, .radius = 1};
  auto my_mat = csg::Mulmatrix2D({{1, 0}}, {{0, 1}});
  my_mat.translation = {2, 0};
  my_mat.group.objs.push_back(my_cir);
  my_rot_ext.group.objs.push_back(my_mat);

  auto my_tree = std::make_shared<csg::Tree>();
  my_tree->top.objs.push_back(my_rot_ext);
  csg::CsgIF my_levelset(my_tree);

  SECTION("Outside") {
    CHECK_FALSE(0 < my_levelset(0, 0, 0));
    CHECK_FALSE(0 < my_levelset(0.9, 0, 0));
    CHECK_FALSE(0 < my_levelset(-0.9, 0, 0));
    CHECK_FALSE(0 < my_levelset(2, 0, 1.1));
    CHECK_FALSE(0 < my_levelset(-2, 0, -1.1));
  }
  SECTION("Inside") {
    CHECK(0 < my_levelset(2, 0, 0));
    CHECK(0 < my_levelset(1.1, 0, 0));
    CHECK(0 < my_levelset(-1.1, 0, 0));
    CHECK(0 < my_levelset(2, 0, 0.9));
    CHECK(0 < my_levelset(-2, 0, -0.9));
  }
}

} // namespace
