#include <sstream>

#include "solver.hpp"
#include "solver_impl.hpp"

namespace solver {

void add_missing_msg(std::string key, std::vector<InputsMessage> &messages,
                     Array defaults) {
  std::ostringstream oss;

  oss << key << " is missing; using defaults: ";
  if (auto sarray = std::get_if<StringArray>(&defaults)) {
    for (auto def : *sarray) {
      oss << def << " ";
    }
  } else if (auto narray = std::get_if<NumberArray>(&defaults)) {
    for (auto def : *narray) {
      oss << def << " ";
    }
  }
  messages.push_back(InputsMessage{oss.str()});
}

void add_msg(std::string key, std::vector<InputsMessage> &messages,
             Array defaults) {
  std::ostringstream oss;
  std::ostringstream defs;
  int size = 0;

  if (auto sarray = std::get_if<StringArray>(&defaults)) {
    size = sarray->size();
    for (auto def : *sarray) {
      defs << def << " ";
    }
  } else if (auto narray = std::get_if<NumberArray>(&defaults)) {
    size = narray->size();
    for (auto def : *narray) {
      defs << def << " ";
    }
  }

  if (size == 1) {
    oss << key << " should be a scalar; using defaults: ";
  } else {
    oss << key << " should have " << size << " elements; using defaults";
  }
  oss << defs.str();
  messages.push_back(InputsMessage{oss.str()});
}

std::pair<solver::SolverSettings, std::vector<InputsMessage>>
make_solver(solver::InputInfo ii) {
  solver::SolverSettings ss;
  auto [geometry, msgs] = make_geometry(ii);
  auto [mesh, msgs2] = make_mesh(ii);
  auto [time, msgs3] = make_time(ii);
  ss.geometry = geometry;
  ss.mesh = mesh;
  ss.time = time;
  std::vector<InputsMessage> messages;
  for (auto msgz :
       std::vector<std::vector<InputsMessage>>{msgs, msgs2, msgs3}) {
    for (auto msg : msgz) {
      messages.push_back(msg);
    }
  }
  return std::make_pair(ss, messages);
}

std::string serialize(solver::SolverSettings settings) {

  std::stringstream buffer;

  auto [xx, yy, zz] = settings.geometry.axes;
  buffer << std::endl;
  buffer << "  geometry.is_periodic = " << xx.periodic << " " << yy.periodic
         << " " << zz.periodic << std::endl;
  buffer << "  geometry.prob_lo     = " << xx.low << " " << yy.low << " "
         << zz.low << std::endl;
  buffer << "  geometry.prob_hi     = " << xx.high << " " << yy.high << " "
         << zz.high << std::endl;

  auto [mx, my, mz] = settings.mesh.axes;
  buffer << "  amr.n_cell      = " << mx.n_cell << " " << my.n_cell << " "
         << mz.n_cell;
  return buffer.str();
}

} // namespace solver
